#! /usr/bin/python
# -*- coding: utf-8 -*-

"""
Class to read BST NenuFAR data
        by A. Loh
"""

__author__ = 'Alan Loh'
__copyright__ = 'Copyright 2018, nenupy'
__credits__ = ['Alan Loh']
__license__ = 'MIT'
__version__ = '0.1.0'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'
__status__ = 'Production'
__all__ = ['BST']


import os
import sys
import glob
import numpy as np

from matplotlib.colors import LogNorm
from matplotlib import pyplot as plt

from astropy.io import fits
from astropy.time import Time, TimeDelta


class BST(object):
    def __init__(self, obsfile):
        self.obsfile = obsfile
        self.abeam   = 0
        self.dbeam   = 0
        self.freq    = 50
        self.polar   = 'nw'

        self._attrlist = ['polar', 'freq', 'time', 'abeam', 'dbeam']

    def __str__(self):
        toprint  = '\t=== Class BST of nenupy ===\n'
        toprint += '\tList of all current attributes:\n'
        for att in dir(self):
            avoid = ['data', 'elana', 'azana', 'azdig', 'eldig', 'maposition', 'marotation', 'delays', 'plot', 'save', 'select']
            if (not att.startswith('_')) & (not any(x.isupper() for x in att)) & (att not in avoid):
                toprint += "%s: %s\n"%(att, getattr(self, att))
        return toprint


    def __repr__(self):
        return '<BST object: obsfile={}>'.format(self.obsfile)

    # ================================================================= #
    # ======================== Getter / Setter ======================== #
    @property
    def obsfile(self):
        """ BST observation file,
            obsfile set to 'fake' will produce a custom fake BST observation
        """
        return self._obsfile
    @obsfile.setter
    def obsfile(self, o):
        if o == 'fake':
            self._createFakeObs()
            self._obsfile = o
            return
        
        if (o is None) or (o == '') or (isinstance(o, str)) and (not os.path.isfile(o)):
            # Look at the current/specified directory
            if o is None:
                return#o = ''
            if os.path.isdir(o):
                _opath = os.path.abspath(o)
                bstfiles = glob.glob( os.path.join(_opath, '*BST.fits') )
            else:
                bstfiles = glob.glob('*BST.fits')
            if len(bstfiles) == 0:
                raise IOError("\n\t=== No BST fits file in current directory, specify the file to read. ===")
            elif len(bstfiles) == 1:
                o = os.path.abspath(bstfiles[0])
            else:
                raise IOError("\n\t=== Multiple BST files are not handled yet ===")
        else:
            if not os.path.isfile(o):
                raise IOError("\t=== File {} not found. ===".format(os.path.abspath(o)))
            else:
                o = os.path.abspath(o) 
        self._obsfile = o

        if not self._isBST():
            raise ValueError("\t=== Files might not be a BST observaiton ===")
        else:
            self._readBST()

        return

    @property
    def abeam(self):
        """ Index of the selected analogic beam
        """
        return self._abeam
    @abeam.setter
    def abeam(self, a):
        if (a is None) or (a==-1):
            return
        elif a not in self.abeams:
            print("\n\t=== WARNING: available AnaBeams are {} ===".format(self.abeams))
            self._abeam = 0
        else:
            self._abeam = a
        self._dbeam = np.arange(self._digi2ana.size)[ self._digi2ana == self._abeam ][0]
        return

    @property
    def dbeam(self):
        """ Index of the selected numeric beam.
        """
        return self._dbeam
    @dbeam.setter
    def dbeam(self, b):
        if (b is None) or (b==-1):
            return
        elif b not in self.dbeams:
            print("\n\t=== WARNING: available DigiBeams are {} ===".format(self.dbeams))
            self._dbeam = 0
        else:
            self._dbeam = b
        freqs = self._freqs[self._dbeam]
        self.freqmin = freqs[freqs > 0.].min()
        self.freqmax = freqs[freqs > 0.].max()
        self._abeam  = self._digi2ana[self._dbeam] # change anabeam automatically
        return

    @property
    def freq(self):
        """ Frequency selection in MHz
            Example:
            self.freq = 50
            self.freq = [20, 30]
        """
        return self._freq
    @freq.setter
    def freq(self, f):
        if isinstance(f, list):
            if len(f) != 2:
                raise ValueError("\n\t=== 'freq' should be a size 2 array ===")
            elif not isinstance(f[0], (np.float32, np.float64, int, float)):
                raise ValueError("\n\t=== 'freq' Syntax error ===")
            elif not isinstance(f[1], (np.float32, np.float64, int, float)):
                raise ValueError("\n\t=== 'freq' Syntax error ===")
            else:
                self._freq = f
        elif isinstance(f, (np.float32, np.float64, int, float)):
            self._freq = f
        else:
            raise ValueError("\n\t=== 'freq' Syntax error ===")
        # if isinstance(f, list):
        #     self._freq    = [0, 0]
        #     self._freq[0] = min(self._freqs[self.dbeam], key=lambda x: np.abs(x-f[0]))
        #     self._freq[1] = min(self._freqs[self.dbeam], key=lambda x: np.abs(x-f[1]))
        # else:
        #     self._freq    = min(self._freqs[self.dbeam], key=lambda x: np.abs(x-f))  
        return 

    @property
    def polar(self):
        """ Polarization selection ('NW' or 'NE')
        """
        return self._polar
    @polar.setter
    def polar(self, p):
        if not isinstance(p, str):
            self._polar = 'NW'
        elif p.upper() not in self._pols:
            print("\n\t=== WARNING: Polarization selection {} not recognized ===".format(p))
            self._polar = 'NW'
        else:
            self._polar = p.upper()

    @property
    def time(self):
        """ Time selection
        """
        return self._time
    @time.setter
    def time(self, t):
        halfsec = TimeDelta(0.5, format='sec')
        if isinstance(t, list):
            if not isinstance(t[0], Time):
                t[0] = Time(t[0])
                
            if not isinstance(t[1], Time):
                t[1] = Time(t[1])
            assert self.obstart-halfsec <= t[0] <= self.obstop+halfsec, 'time[0] outside time range'
            assert self.obstart-halfsec <= t[1] <= self.obstop+halfsec, 'time[1] outside time range'
            assert t[0] < t[1], 'time[0] > time[1]!'
            self._time = t
        else:
            if not isinstance(t, Time):
                t = Time(t)
            assert self.obstart-halfsec <= t <= self.obstop+halfsec, 'time outside time range'
            self._time = t
            return

    # ------ Specific getters ------ #
    @property
    def ma(self):
        """ Mini-Arrays used
        """
        self._ma = self._malist[ self.abeam ]
        return self._ma

    @property
    def marotation(self):
        """ Mini-array rotation in degrees
        """
        self._marotation = self._marot[ np.searchsorted(self._maon, self.ma) ]
        return self._marotation

    @property
    def maposition(self):
        """ Lambert93 (x, y, z) Mini-Array positions
        """
        self._maposition = self._mapos[ np.searchsorted(self._maon, self.ma) ]
        return self._maposition

    @property
    def delays(self):
        """ Delays in nsec
        """
        self._delays = self._nsdelays[ self.ma ]
        return self._delays

    @property
    def azana(self):
        """ Azimuth during the analogic pointing self.abeam
        """
        az = self._azlistana[ self._pointana==self.abeam ]
        if az.size == 1:
            az = az[0]
        else:
            anatimes = self._pointanat[self.abeam][ self._pointana==self.abeam ]
            if isinstance(self.time, list):
                tmask = np.squeeze((anatimes >= self.time[0]) & (anatimes <= self.time[1]))
                az = az[tmask]
            else:
                # az = az[ np.argmin(np.abs( (anatimes - self.time).sec )) ] one should take the previous value not the closest!
                az = az[(anatimes <= self.time)][-1]
        return az 

    @property
    def elana(self):
        """ Elevation during the analogic pointing self.abeam
        """
        el = self._ellistana[ self._pointana==self.abeam ]
        if el.size == 1:
            el = el[0]
        else:
            anatimes = self._pointanat[self.abeam][ self._pointana==self.abeam ]
            if isinstance(self.time, list):
                tmask = np.squeeze((anatimes >= self.time[0]) & (anatimes <= self.time[1]))
                # el = self._ellistana[tmask]
                el = el[tmask]
            else:
                # el = self._ellistana[ np.argmin(np.abs( (anatimes - self.time).sec )) ] one should take the previous value not the closest!
                el = el[(anatimes <= self.time)][-1]
        return el

    @property
    def azdig(self):
        """ Azimuth during the numeric pointing self.dbeam
        """
        az = self._azlistdig[ self._pointdig==self.dbeam ]
        if az.size == 1:
            az = az[0]
        else:
            digitimes = self._pointdigt[ self._pointdig==self.dbeam ]
            if isinstance(self.time, list):
                tmask = np.squeeze((digitimes >= self.time[0]) & (digitimes <= self.time[1]))
                az = az[tmask]
            else:
                # az = az[ np.argmin(np.abs( (digitimes - self.time).sec )) ] one should take the previous value not the closest!
                az = az[(digitimes <= self.time)][-1]
        return az

    @property
    def eldig(self):
        """ Elevation during the numeric pointing self.dbeam
        """ 
        el = self._ellistdig[ self._pointdig==self.dbeam ]
        if el.size == 1:
            el = el[0]
        else:
            digitimes = self._pointdigt[ self._pointdig==self.dbeam ]
            if isinstance(self.time, list):
                tmask = np.squeeze((digitimes >= self.time[0]) & (digitimes <= self.time[1]))
                el = el[tmask]
            else:
                # el = el[ np.argmin(np.abs( (digitimes - self.time).sec )) ] # one should take the previous value not the closest!
                el = el[(digitimes <= self.time)][-1]
        return el

    @ property
    def angles(self):
        """ Pointed angle
        """
        if self.type == 'tracking':
            from nenupy.astro import toRadec
            digitimes = self._pointdigt[ self._pointdig==self.dbeam ]
            if isinstance(self.time, list):
                tmask = np.squeeze((digitimes >= self.time[0]) & (digitimes <= self.time[1]))
                tt = digitimes[tmask][0]
            else:
                tt = digitimes[(digitimes <= self.time)][-1]
            radec = toRadec(source=(self.azdig[0], self.eldig[0]),
                time=tt,
                loc='NenuFAR')
            return (radec.ra.deg, radec.dec.deg, 'J2000')
        else:
            self.type = 'transit'
            return (self.azdig, self.eldig, 'ALTAZ')


    # ================================================================= #
    # =========================== Methods ============================= #
    def select(self, **kwargs):
        """ Make the data selection

            Parameters
            ----------
            kwargs : {freq, polar, time, abeam, dbeam}
                Keyword arguments

            Returns
            -------
            self.d : np.ndarray
                Data selected
            self.f : np.ndarray
                Frequency selected
            self.t : np.ndarray
                Time selected
        """
        self._evalkwargs(kwargs)

        # ------ load data only once ------ #
        if not hasattr(self, '_dataall'):
            with fits.open(self.obsfile, mode='readonly', ignore_missing_end=True, memmap=True) as fitsfile:
               self._timeall = Time( fitsfile[7].data['jd'], format='jd' ) 
               self._dataall = fitsfile[7].data['data'] 

        # ------ polarization ------ #
        mask_pol = (self._pols == self.polar)

        # ------ frequency ------ #
        if isinstance(self.freq, list):
            freqi = min(self._freqs[self.dbeam], key=lambda x: np.abs(x-self.freq[0]))
            freqf = min(self._freqs[self.dbeam], key=lambda x: np.abs(x-self.freq[1]))
            #mask_fre = (self._freqs[self.dbeam] > 0.) & (self._freqs[self.dbeam] >= self.freq[0]) & (self._freqs[self.dbeam] <= self.freq[-1])
            mask_fre = (self._freqs[self.dbeam] > 0.) & (self._freqs[self.dbeam] >= freqi) & (self._freqs[self.dbeam] <= freqf)
        else:
            mask_fre = (self._freqs[self.dbeam] == min(self._freqs[self.dbeam], key=lambda x: np.abs(x-self.freq)))
        mask_freq = np.roll(mask_fre, self._bletlist[self.dbeam][0])

        # ------ time ------ #
        digitimes = self._pointdigt[ self._pointdig==self.dbeam ]
        if digitimes.size == 1:
            digiti   = digitimes[0]
            digitf   = digitimes[0] + self._ddeltat[0]
        else:
            digiti   = digitimes[0]
            digitf   = digitimes[-1]
        mask_tim = (self._timeall >= digiti) & (self._timeall <= digitf)
        if isinstance(self.time, list):
            mask_tim = mask_tim & (self._timeall >= self.time[0]) & (self._timeall <= self.time[1])
        else:
            mask_tim = mask_tim & (self._timeall.mjd == min(self._timeall.mjd, key=lambda x: np.abs(x-self.time.mjd)) )
       
        # ------ selected data ------ #
        d = np.squeeze( self._dataall[ np.ix_(mask_tim, mask_pol, mask_freq) ] )
        t = self._timeall[ mask_tim ]
        f = self._freqs[self.dbeam][ mask_fre ]
        self.data = {'amp': d, 'freq': f, 'time': t}
        return

    def plot(self, db=True, **kwargs):
        """ Plot the data

            .. note::
                This function uses the same keyword argument as :func:`select`.

            Args:
                * db : `bool`
                    We all know what foo does.

            Kwargs:
                * freq : `float` or `list`
                    The frequency in MHz
                
                * polar : `str`
                    The polarization (`'NE'` or `'NW'`)
                                
                * time : `str`
                    The time

                * abeam : `int`
                
                * dbeam : `int`
            
            Example:
                >>> plot(db=True, freq=50, polar='NE')
        """
        self.select(**kwargs)

        plotkwargs = {key: value for (key, value) in kwargs.items() if key not in self._attrlist}

        if self.data['freq'].size == 1:
            # ------ Light curve ------ #
            xtime = (self.data['time'] - self.data['time'][0]).sec / 60
            if db:
                plt.plot(xtime, 10.*np.log10(self.data['amp']), **plotkwargs)
                plt.ylabel('dB')
            else:
                plt.plot(xtime, self.data['amp'], **plotkwargs)
                plt.ylabel('Amplitude')
            plt.xlabel('Time (min since {})'.format(self.data['time'][0].iso))
            plt.title('f={:3.2f} MHz, pol={}, abeam={}, dbeam={}'.format(self.data['freq'][0], self.polar, self.abeam, self.dbeam))

        elif self.data['time'].size == 1:
            # ------ Spectrum ------ #
            if db:
                plt.plot(self.data['freq'], 10.*np.log10(self.data['amp']), **plotkwargs)
                plt.ylabel('dB')
            else:
                plt.plot(self.data['freq'], self.data['amp'], **plotkwargs)
                plt.ylabel('Amplitude')
            plt.xlabel('Frequency (MHz)')
            plt.title('t={}, pol={}, abeam={}, dbeam={}'.format(self.time.iso, self.polar, self.abeam, self.dbeam))

        elif (self.data['freq'].size > 1) & (self.data['time'].size > 1):
            # ------ Dynamic spectrum ------ #
            xtime = (self.data['time'] - self.data['time'][0]).sec / 60
            vmin, vmax = np.percentile(self.data['amp'], [5, 99])
            cmap = 'bone'
            for key, value in plotkwargs.items():
                if key == 'cmap': cmap = value
                if key == 'vmin': vmax = value
                if key == 'vmax': vmin = value 
            fig = plt.figure()
            ax  = fig.add_subplot(111)
            normcb = LogNorm(vmin=vmin, vmax=vmax)
            spec   = ax.pcolormesh(xtime, self.data['freq'], self.data['amp'].T, cmap=cmap, norm=normcb)
            plt.colorbar(spec)
            ax.axis( [xtime.min(), xtime.max(), self.data['freq'].min(), self.data['freq'].max()] )
            plt.xlabel('Time (min since {})'.format(self.data['time'][0].iso))
            plt.ylabel('Frequency (MHz)')
            plt.title('pol={}, abeam={}, dbeam={}'.format(self.polar, self.abeam, self.dbeam))

        else:
            raise ValueError("\n\t=== ERROR: Plot nature not understood ===")

        if hasattr(self, 'src'):
            if self.type == 'transit':
                try:
                    from ..astro import getTransit
                    transit = getTransit(source=self.src, time=self.obstart, loc='Nenufar', az=self.azdig)
                    transit = (transit - self.data['time'][0]).sec / 60.
                    plt.axvline(x=transit, color='black', linestyle='-.', linewidth=1)
                except:
                    pass

        plt.show()
        plt.close('all')
        return

    def save(self, savefile=None, **kwargs):
        """ Save the data
        """
        if savefile is None:
            savefile = self.obsname + '_data.fits'
        else:
            if not savefile.endswith('.fits'):
                raise ValueError("\n\t=== It should be a FITS ===")
        self.select(**kwargs)

        prihdr = fits.Header()
        prihdr.set('OBS', self.obsname)
        prihdr.set('FREQ', str(self.freq))
        prihdr.set('TIME', str(self.time))
        prihdr.set('POLAR', self.polar)
        #prihdr.set('MINI-ARR', str(self.ma))
        datahdu = fits.PrimaryHDU(self.d.T, header=prihdr)
        freqhdu = fits.BinTableHDU.from_columns( [fits.Column(name='frequency', format='D', array=self.data['freq'])] )
        timehdu = fits.BinTableHDU.from_columns( [fits.Column(name='mjd', format='D', array=self.data['time'].mjd)] )
        hdulist = fits.HDUList([datahdu, freqhdu, timehdu])
        hdulist.writeto(savefile, overwrite=True)
        return

    def mean(self, axis='freq'):
        """
        """
        if axis.lower() == 'freq':
            if self.data['freq'].size > 1:
                self.data['amp'] = np.mean(self.data['amp'], axis=1)
                self.data['freq'] = np.mean(self.data['freq'])
        elif axis.lower() == 'time':
            if self.data['time'].size > 1:
                self.data['amp'] = np.mean(self.data['amp'], axis=0)
                self.data['time'] = Time(np.mean(self.data['time'].mjd), format='mjd')
        else:
            print('Axis {} not in [freq, time]'.format(axis))
        return

    def from_transit(self, source, unit='sec'):
        """ Redefine the time with respect to the transit time
        """
        if self.type == 'transit':
            # try:
            from ..astro import getTransit
            transit = getTransit(source=source, time=self.obstart, loc='Nenufar', az=self.azdig)
            if unit.lower() == 'sec':
                self.data['time'] = (self.data['time'] - transit).sec
            elif unit.lower() == 'min':
                self.data['time'] = (self.data['time'] - transit).sec / 60.
            elif unit.lower() == 'hour':
                self.data['time'] = (self.data['time'] - transit).sec / 3600.
            else:
                print('Accepted values are [sec, min, hour]')
        #     except:
        #         pass
        return

    # ================================================================= #
    # =========================== Internal ============================ #
    def _isBST(self):
        """ Check that self.obsfile is a proper BST observation
        """
        isBST = True
        with fits.open(self.obsfile, mode='readonly', ignore_missing_end=True, memmap=True) as f:
            if f[0].header['OBJECT'] != 'beamlet Statistics':
                isBST = False
            else:
                pass
        return isBST

    def _readBST(self):
        """ Read BST fits files and fill the class attributes
        """
        self.obsname = os.path.basename( self.obsfile )[:-9]

        with fits.open(self.obsfile, mode='readonly', ignore_missing_end=True, memmap=True) as f:
            head      = f[0].header
            setup_ins = f[1].data
            setup_obs = f[2].data
            setup_ana = f[3].data
            setup_bea = f[4].data
            setup_pan = f[5].data
            setup_pbe = f[6].data

        try:
            self.name = setup_obs['name']
        except:
            self.name = ''

        self.obstart  = Time( head['DATE-OBS'] + 'T' + head['TIME-OBS'] )
        self.obstop   = Time( head['DATE-END'] + 'T' + head['TIME-END'] )
        self.time     = [self.obstart.copy(), self.obstop.copy()]
        self.exposure = self.obstop - self.obstart

        self._maon  = np.squeeze( setup_ins['noMROn'] )
        self._marot = np.squeeze( setup_ins['rotation'] )
        self._mapos = np.squeeze( setup_ins['noPosition'] )
        self._mapos = self._mapos.reshape( int(self._mapos.size/3), 3 )
        self._pols  = np.squeeze( setup_ins['spol'] )
        try:
            self._nsdelays = np.squeeze(setup_ins['delay'])[::2]#[ self._maon ]
        except:
            self._nsdelays = np.zeros( self._maon.max() )

        self.abeams     = setup_ana['NoAnaBeam']
        self._antlist   = np.array( [ np.array(eval(i)) - 1 for i in setup_ana['Antlist'] ])
        self._adeltat   = TimeDelta(setup_ana['Duration'], format='sec')
        self._malist    = np.array( [mrs[0:setup_ana['nbMRUsed'][i]] for i, mrs in enumerate(setup_ana['MRList'])] )

        self.dbeams     = setup_bea['noBeam']
        self._digi2ana  = setup_bea['NoAnaBeam']
        self._bletlist  = setup_bea['BeamletList']
        self._freqs     = setup_bea['freqList']
        self._ddeltat   = TimeDelta(setup_bea['Duration'], format='sec')

        self._pointana  = setup_pan['noAnaBeam']
        self._azlistana = setup_pan['AZ']
        self._ellistana = setup_pan['EL']
        self._pointanat = Time(np.array([setup_pan['timestamp'][self._pointana==i] for i in self.abeams]))
        
        self._pointdig  = setup_pbe['noBeam']
        self._azlistdig = setup_pbe['AZ']
        self._ellistdig = setup_pbe['EL']
        self._pointdigt = Time(setup_pbe['timestamp'])

        if self._pointanat.shape[1] != 1:
            self.type = 'tracking'
        else:
            self.type = 'transit'
        return

    def _createFakeObs(self):
        return

    def _evalkwargs(self, kwargs):
        for key, value in kwargs.items():
            if   key == 'polar': self.polar = value
            elif key == 'freq':  self.freq  = value
            elif key == 'time':  self.time  = value
            elif key == 'abeam': self.abeam = value
            elif key == 'dbeam': self.dbeam = value
            elif key == 'src':   self.src   = value
            else:
                pass
        return
    



#! /usr/bin/python
# -*- coding: utf-8 -*-

import numpy as np
from matplotlib import pyplot as plt
import matplotlib.dates as mdates
from scipy.interpolate import interp1d
import shutil
import os

from astropy.time import Time, TimeDelta
from astropy import units as u
from astropy import coordinates as coord

from .ateam import Source

__author__ = ['Alan Loh']
__copyright__ = 'Copyright 2018, celespy'
__credits__ = ['Alan Loh']
__license__ = 'MIT'
__version__ = '0.0.1'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'
__status__ = 'WIP'
__all__ = [ 'getLoc', 'getSrc', 'getAltaz', 'getTime',
            'getTransit', 'getSep', 'riseTime', 'setTime',
            'meridianTime', 'plotElevation', 'plotElevations',
            'plotSeparation', 'toRadec']


# =================================================================================== #
# ------------------------------------- getLoc -------------------------------------- #
# =================================================================================== #
def getLoc(loc, unit='deg'):
    """ Get the a location

        Parameters
        ----------
        * **loc** : tuple or str
            Location query (can be a tuple or a string).
            If tuple, expects (longitude, latitude).
            If string, expects an address, typically `'Paris, France'` and queries Google.
        * **unit** : str, optional
            If `loc` is a tuple of lon/lat, unit can be either `'deg'` or `'rad'`

        Returns
        -------
        * **loc** : ``astropy.coord.EarthLocation``
            EarthLocation object
    """
    if isinstance(loc, str):
        if loc.lower() == 'nenufar':
            old_loc = coord.EarthLocation(lat=47.376511*u.deg, lon=2.1924002*u.deg)
            loc = coord.EarthLocation(lat=47.375944*u.deg, lon=2.193361*u.deg, height=136.195*u.m)
            # latitude=47.375944            ; 47deg22'33.39729" N
            # longitude=2.193361         ; 2deg11'36.09956" E
            # altitude=136.195             ; 136,195 m
            # X_LAMBERT93=639133.971
            # Y_LAMBERT93=6697600.364
            # Z_LAMBERT93=182.096
        else:
            # if not USE_GEOPY:
            #     loc = coord.EarthLocation.of_address(loc)
            # else:
            #     geoloc = Nominatim(user_agent='my-application')
            #     gloc = geoloc.geocode(loc)
            #     loc = coord.EarthLocation(lat=gloc.latitude*u.deg, lon=gloc.longitude*u.deg)
            print('not implemented')
    elif isinstance(loc, tuple):
        assert len(loc)==2, 'Only length 2 tuple is understood.'
        if not unit.lower() == 'deg':
            loc = np.degrees(loc)
        loc = coord.EarthLocation(lat=loc[0]*u.deg, lon=loc[1]*u.deg)
    elif isinstance(loc, coord.EarthLocation):
        pass
    else:
        raise ValueError('loc must be aither a tuple: (lon, lat) in degrees or a string adress.')

    return loc


# =================================================================================== #
# ------------------------------------- getSrc -------------------------------------- #
# =================================================================================== #
def getSrc(source, time=None, loc=None, unit='deg'):
    """ Find a specific source
        
        Parameters
        ----------
        * **source** : str, tuple
            Soure name / cooodinates to be resolved
        * **time** : str, number or ``astropy.time.Time``
            The UTC time at which the conversion should be computed
        * **loc** : str or tuple
            The location at which the conversion should be computed
        * **unit** : str, optional
            If `source` is a tuple of ra/dec, unit can be either `'deg'` or `'rad'`

        Returns
        -------
        * **src** : ``astropy.coord.SkyCoord``
            SkyCoord object
    """
    try:
        ateam = Source(source=source)
        src = ateam.source
    except:
        if isinstance(source, str):
            if source.lower() in ['sun', 'moon', 'jupiter', 'saturn', 'mars', 'venus']:
                time = getTime(time)
                loc  = getLoc(loc)
                try:
                    with coord.solar_system_ephemeris.set('builtin'):
                        src = coord.get_body(source, time, loc)
                except:
                    src = None
            else:
                src = coord.SkyCoord.from_name(source)
        elif isinstance(source, tuple):
            assert len(source)==2, 'Only length 2 tuple is understood.'
            if not unit.lower() == 'deg':
                ra, dec = np.degrees(source)
            else:
                ra, dec = source
            src = coord.SkyCoord(ra*u.deg, dec*u.deg, frame='icrs')
        elif isinstance(source, coord.SkyCoord):
            src = source
        else:
            raise ValueError('source parameter not understood.')

    return src


# =================================================================================== #
# ------------------------------------ getAltaz ------------------------------------- #
# =================================================================================== #
def getAltaz(source, time, loc, unit='deg'):
    """ Convert Equatorial coordinates (in deg) to the Horizontal (alt-az) system

        Parameters
        ----------
        * **source** : str, tuple or ``astropy.coord.SkyCoord``
            The astrophysical position/source to convert into Alt Az
        * **time** : str, number or ``astropy.time.Time``
            The UTC time at which the conversion should be computed
        * **loc** : str or tuple
            The location at which the conversion should be computed
        * **unit** : str, optional
            If `source` is a tuple of ra/dec, unit can be either `'deg'` or `'rad'`

        Returns
        -------
        * **altaz** : ``astropy.coord.SkyCoord``
            SkyCoord object in alt az coordinates
    """
    if isinstance(source, str):
        source = getSrc(source=source, time=time, loc=loc, unit=unit)
        ra, dec = source.ra.deg, source.dec.deg
    elif isinstance(source, tuple):
        assert len(source)==2, 'Only length 2 tuple is understood.'
        ra, dec = source
        if not unit.lower() == 'deg':
            ra, dec = np.degrees(ra, dec)
    elif isinstance(source, coord.SkyCoord):
        ra, dec = source.ra.deg, source.dec.deg
    else:
        raise ValueError('source parameter not understood.')

    time  = getTime(time)
    loc   = getLoc(loc) 
    radec = coord.SkyCoord(ra*u.deg, dec*u.deg, frame='icrs')
    frame = coord.AltAz(obstime=time, location=loc)
    altaz = radec.transform_to(frame)

    return altaz


# =================================================================================== #
# ------------------------------------ toRadec -------------------------------------- #
# =================================================================================== #
def toRadec(source, time, loc, unit='deg'):
    """ Convert Equatorial coordinates (in deg) to the Horizontal (alt-az) system

        Parameters
        ----------
        * **source** : tuple
            (az, alt)
        * **time** : str, number or ``astropy.time.Time``
            The UTC time at which the conversion should be computed
        * **loc** : str or tuple
            The location at which the conversion should be computed
        * **unit** : str, optional
            If `source` is a tuple of ra/dec, unit can be either `'deg'` or `'rad'`

        Returns
        -------
        * **altaz** : ``astropy.coord.SkyCoord``
            SkyCoord object in alt az coordinates
    """
    if isinstance(source, tuple):
        az, alt = source
        if not unit.lower() == 'deg':
            az = np.degrees(az)
            alt = np.degrees(alt)
    else:
        raise ValueError('source parameter not understood.')
    time  = getTime(time)
    loc   = getLoc(loc) 
    azel  = coord.SkyCoord(alt=alt*u.deg, az=az*u.deg, obstime=time, location=loc, frame='altaz')
    
    return azel.fk5


# =================================================================================== #
# ------------------------------------- getTime ------------------------------------- #
# =================================================================================== #
def getTime(time, unit=None):
    """ Convert time to a astropy.Time object

        Parameters
        ----------
        * **time** : ``astropy.time.Time``, number or str
            Time to parse (`'YYYY-MM-DD hh:mm:ss'`)
            Understands also `'now'`, `'tomorrow'`, `'yesterday'`
        * **unit** : str, optional
            If `time` is a number, unit can be `'jd'` or `'mjd'`

        Returns
        -------
        * **time** : ``astropy.time.Time``
            Time object
    """
    if isinstance(time, Time):
        pass
    elif isinstance(time, str):
        if time.lower() == 'now':
            time = Time.now()
        elif time.lower() == 'tomorrow':
            time = Time.now() + TimeDelta(1, format='jd')
        elif time.lower() == 'yesterday':
            time = Time.now() - TimeDelta(1, format='jd')
        else:
            time = Time(time)
    elif isinstance(time, (int, float, np.number)):
        if unit is None:
            unit = 'jd'
        time = Time(time, format=unit.lower())
    else:
        raise ValueError('time parameter not understood.')

    return time


# =================================================================================== #
# ----------------------------------- getTransit ------------------------------------ #
# =================================================================================== #
def getTransit(source, time, loc, az=None, el=None, unit='deg', way='rise', t_step=1800):
    """ Get the *next* transit time for a particular source/position
        at a given position (obs site) since a given time.
        Transit may be defined by the next crossing of a particular azimuth *or* elevation.

        Parameters
        ----------
        * **source** : str, tuple or ``astropy.coord.SkyCoord``
            The astrophysical position/source to convert into Alt Az
        * **time** : str, number or Time
            The UTC time at which the conversion should be computed
        * **loc** : str or tuple
            The location at which the conversion should be computed
        * **az** : number, optional
            Azimuth at which look for the transit (degrees)
        * **el** : number, optional
            Elevation at which look for the transit (degrees)
        * **unit** : str, optional
            If `source` is a tuple of ra/dec, unit can be either `'deg'` or `'rad'`
        * **way** : str, optional
            If `el` corssing is defined, precise the direction `'rise'` or `set`
        * **t_step** : float, optional
            Time step to look for transit

        Returns
        -------
        * **transit_time** : ``astropy.time.Time`
            Transit time (precision : +/- 2.5 sec)

        Example
        -------
        To get the meridian transit time of Cyg A
            >>> getTransit('Cyg A', 'now', 'nenufar', az=180)
    """
    time   = getTime(time)
    loc    = getLoc(loc)
    source = getSrc(source=source, time=time, loc=loc, unit=unit)

    assert (way.lower() == 'rise') or (way.lower() == 'set'), 'way must be either rise or set'

    def find_time_interval(start, dt, source, az, el):
        while start <= start + dt:
            if (az is not None) & (el is None):
                c1 = getAltaz(source, start, loc).az.deg
                c2 = getAltaz(source, start+dt, loc).az.deg
                if source.dec.deg > loc.lat.deg:
                    deltaaz = np.degrees(np.arccos((90-source.dec.deg)/loc.lat.deg)) # az span
                    # circumpolar
                    if (0. < az <= 0.+deltaaz):
                        if   (270. <= c1 < 360.) & (270. <= c2 < 360.):
                            pass
                        elif (0. <= c1 < 90.) & (0. <= c2 < 90.):
                            if (c2 <= az <= c1):
                                break
                        else:
                            if (270. <= c2 <= 360.) & (c1 >= az):
                                break
                    elif (az == 0.) or (az == 360.):
                        if (c2 - c1 > 180):
                            break
                    elif (360.-deltaaz <= az < 360.):
                        if   (270. <= c1 < 360.) & (270. <= c2 < 360.):
                            if (c2 <= az <= c1):
                                break
                        elif (0. <= c1 < 90.) & (0. <= c2 < 90.):
                            pass
                        else:
                            if (0. <= c1 <= 90.) & (c2 <= az):
                                break
                    else:
                        raise ValueError('Azimuth {} not valid for a circumpolar source, only 0.+/-{}'.format(az, deltaaz))
                    
                else:
                    if (c1 <= az <= c2):
                        break
            elif (el is not None) & (az is None):
                c1 = getAltaz(source, start, loc).alt.deg
                c2 = getAltaz(source, start+dt, loc).alt.deg
                if way.lower() == 'rise':
                    if (c1 <= el <= c2):
                        break
                elif way.lower() == 'set':
                    if (c2 <= el <= c1):
                        break
            else:
                raise ValueError('either az or el parameters should not be None')
            start += dt
        return start

    bigdt = TimeDelta(t_step, format='sec') 
    new_start = find_time_interval(start=time, dt=bigdt, source=source, az=az, el=el)
    middt = TimeDelta(t_step/6., format='sec')
    new_start = find_time_interval(start=new_start, dt=middt, source=source, az=az, el=el)
    smalldt = TimeDelta(t_step/360, format='sec')
    new_start = find_time_interval(start=new_start, dt=smalldt, source=source, az=az, el=el)
    smallestdt = TimeDelta(t_step/1800., format='sec')
    new_start = find_time_interval(start=new_start, dt=smallestdt, source=source, az=az, el=el)

    return new_start + smallestdt/2.


# =================================================================================== #
# ------------------------------------- riseTime ------------------------------------ #
# =================================================================================== #
def riseTime(source, time, loc, unit='deg'):
    """ Compute the rise time of a source
        This is defined as 0 deg elevation crossing.

        Parameters
        ----------
        * **source** : str, tuple or ``astropy.coord.SkyCoord``
            The astrophysical position/source to convert into Alt Az
        * **time** : str, number or Time
            The UTC time at which the conversion should be computed
        * **loc** : str or tuple
            The location at which the conversion should be computed
        * **unit** : str, optional
            If `source` is a tuple of ra/dec, unit can be either `'deg'` or `'rad'`

        Returns
        -------
        * **rising_time** : ``astropy.time.Time`
            Rising time (precision : +/- 2.5 sec)

        Example
        -------
        To get the rising time of Cyg A
            >>> riseTime('Cyg A', 'now', 'nenufar')
    """
    rise = getTransit(source=source,
                        time=time,
                        loc=loc,
                        el=0,
                        unit=unit,
                        way='rise')

    return rise


# =================================================================================== #
# ------------------------------------- setTime ------------------------------------- #
# =================================================================================== #
def setTime(source, time, loc, unit='deg'):
    """ Compute the set time of a source
        This is defined as 0 deg elevation crossing.

        Parameters
        ----------
        * **source** : str, tuple or ``astropy.coord.SkyCoord``
            The astrophysical position/source to convert into Alt Az
        * **time** : str, number or Time
            The UTC time at which the conversion should be computed
        * **loc** : str or tuple
            The location at which the conversion should be computed
        * **unit** : str, optional
            If `source` is a tuple of ra/dec, unit can be either `'deg'` or `'rad'`

        Returns
        -------
        * **setting_time** : ``astropy.time.Time`
            Setting time (precision : +/- 2.5 sec)

        Example
        -------
        To get the setting time of Cyg A
            >>> setTime('Cyg A', 'now', 'nenufar')
    """
    sett = getTransit(source=source,
                        time=time,
                        loc=loc,
                        el=0,
                        unit=unit,
                        way='set')

    return sett


# =================================================================================== #
# ----------------------------------- meridianTime ---------------------------------- #
# =================================================================================== #
def meridianTime(source, time, loc, unit='deg'):
    """ Compute the set time of a source
        This is defined as 0 deg elevation crossing.

        Parameters
        ----------
        * **source** : str, tuple or ``astropy.coord.SkyCoord``
            The astrophysical position/source to convert into Alt Az
        * **time** : str, number or Time
            The UTC time at which the conversion should be computed
        * **loc** : str or tuple
            The location at which the conversion should be computed
        * **unit** : str, optional
            If `source` is a tuple of ra/dec, unit can be either `'deg'` or `'rad'`

        Returns
        -------
        * **meridian_time** : ``astropy.time.Time`
            Meridian transit time (precision : +/- 2.5 sec)

        Example
        -------
        To get the setting time of Cyg A
            >>> meridianTime('Cyg A', 'now', 'nenufar')
    """
    meridian = getTransit(source=source,
                        time=time,
                        loc=loc,
                        az=180,
                        unit=unit)

    return meridian


# =================================================================================== #
# ------------------------------------- getSep -------------------------------------- #
# =================================================================================== #
def getSep(source1, source2, time=None, loc=None, unit='deg'):
    """ Get the angular separation between two sources

        Parameters
        ----------
        * **source1** : str, tuple
            Soure name / cooodinates to be resolved
        * **source2** : str, tuple
            Soure name / cooodinates to be resolved
        * **unit** : str, optional
            If `source` is a tuple of ra/dec, unit can be either `'deg'` or `'rad'`

        Returns
        -------
        * **angular_separation** : ``astropy.coordinates.angles.Angle``
            Angular separation between `source1` and `source2`
    """
    source1 = getSrc(source=source1, time=time, loc=loc, unit=unit)
    source2 = getSrc(source=source2, time=time, loc=loc, unit=unit)

    return source1.separation(source2)


# =================================================================================== #
# ---------------------------------- plotElevation ---------------------------------- #
# =================================================================================== #
def plotElevation(source, time, loc, unit='deg', dt=1800):
    """ Plot the source elevation for a day since `time`

        Parameters
        ----------
        * **source** : str, tuple or ``astropy.coord.SkyCoord``
            The astrophysical position/source to convert into Alt Az
        * **time** : str, number or Time
            The UTC time at which the conversion should be computed
        * **loc** : str or tuple
            The location at which the conversion should be computed
        * **unit** : str, optional
            If `source` is a tuple of ra/dec, unit can be either `'deg'` or `'rad'`
        * **dt** : number, optional
            Time interval for the elevation computation (seconds)
    """
    time   = getTime(time)
    t0     = time.iso.split()[0].replace('-', '/') 
    sname  = source
    lname  = loc 
    loc    = getLoc(loc)
    source = getSrc(source=source, time=time, loc=loc, unit=unit)
    
    # Compute the elevation vs time
    dt    = TimeDelta(dt, format='sec')
    nvals = int((TimeDelta(1, format='jd')/dt).value)
    times = np.zeros( nvals )
    elev  = np.zeros( nvals )
    for i in np.arange(nvals):
        src = getAltaz(source=source, time=time, loc=loc, unit=unit)
        times[i] = time.mjd
        elev[i]  = src.alt.deg
        time += dt

    # Interpolate
    evst  = interp1d(times, elev, kind='cubic')
    times = np.linspace(times[0], times[-1], 1000) 
    elev  = evst(times)

    # Plot
    fig, ax = plt.subplots()
    xtime = mdates.date2num( Time(times, format='mjd').to_datetime() )
    ax.plot( xtime, elev, label='{} @ {}'.format(sname, lname) )
    ax.axhline(0., linestyle='--', color='black', linewidth=1)
    ymin, ymax = ax.get_ylim()
    ax.fill_between( xtime, ymin, ymax,
        where=(elev <= 0.), facecolor='gray', alpha=0.5, step='post')
    ax.fill_between( xtime, ymin, ymax,
        where=(0. <= elev) & (elev <= 20.), facecolor='C3', alpha=0.5, step='mid')
    ax.fill_between( xtime, ymin, ymax,
        where=(20. <= elev) & (elev <= 40.), facecolor='C1', alpha=0.5, step='mid')
    ax.fill_between( xtime, ymin, ymax,
        where=(40. <= elev), facecolor='C2', alpha=0.5, step='mid')
    ax.xaxis_date()
    date_format = mdates.DateFormatter('%H:%M')
    ax.xaxis.set_major_formatter(date_format)
    ax.xaxis.set_minor_formatter(date_format)
    fig.autofmt_xdate()
    ax.legend()
    ax.set_xlabel('Time ({} UTC)'.format(t0))
    ax.set_ylabel('Elevation (degrees)')
    plt.show()


# =================================================================================== #
# ---------------------------------- plotElevations --------------------------------- #
# =================================================================================== #
def plotElevations(sources, time, loc, unit='deg', dt=1800):
    """ Plot the source elevation for a day since `time`

        Parameters
        ----------
        * **source** : str, tuple or ``astropy.coord.SkyCoord``
            The astrophysical position/source to convert into Alt Az
        * **time** : str, number or Time
            The UTC time at which the conversion should be computed
        * **loc** : str or tuple
            The location at which the conversion should be computed
        * **unit** : str, optional
            If `source` is a tuple of ra/dec, unit can be either `'deg'` or `'rad'`
        * **dt** : number, optional
            Time interval for the elevation computation (seconds)
    """
    fig, ax = plt.subplots()
    lname  = loc
    tini = time
    for source in sources:
        time   = getTime(tini)
        t0     = time.iso.split()[0].replace('-', '/') 
        sname  = source
        loc    = getLoc(loc)
        source = getSrc(source=source, time=time, loc=loc, unit=unit)
        
        # Compute the elevation vs time
        dt    = TimeDelta(dt, format='sec')
        nvals = int((TimeDelta(1, format='jd')/dt).value)
        times = np.zeros( nvals )
        elev  = np.zeros( nvals )
        for i in np.arange(nvals):
            src = getAltaz(source=source, time=time, loc=loc, unit=unit)
            times[i] = time.mjd
            elev[i]  = src.alt.deg
            time += dt

        # Interpolate
        evst  = interp1d(times, elev, kind='cubic')
        times = np.linspace(times[0], times[-1], 1000) 
        elev  = evst(times)

        # Plot
        xtime = mdates.date2num( Time(times, format='mjd').to_datetime() )
        ax.plot( xtime, elev, label='{} @ {}'.format(sname, lname) )
        ax.axhline(0., linestyle='--', color='black', linewidth=1)
        ymin, ymax = ax.get_ylim()
    
        ax.xaxis_date()
        date_format = mdates.DateFormatter('%H:%M')
        ax.xaxis.set_major_formatter(date_format)
        ax.xaxis.set_minor_formatter(date_format)
        fig.autofmt_xdate()
        ax.legend()
    ax.set_xlabel('Time ({} UTC)'.format(t0))
    ax.set_ylabel('Elevation (degrees)')
    plt.show()


# =================================================================================== #
# ---------------------------------- plotSeparation --------------------------------- #
# =================================================================================== #
def plotSeparation(sources, time, loc, unit='deg', dt=1800):
    """ Plot the source elevation for a day since `time`

        Parameters
        ----------
        * **source** : str, tuple or ``astropy.coord.SkyCoord``
            The astrophysical position/source to convert into Alt Az
        * **time** : str, number or Time
            The UTC time at which the conversion should be computed
        * **loc** : str or tuple
            The location at which the conversion should be computed
        * **unit** : str, optional
            If `source` is a tuple of ra/dec, unit can be either `'deg'` or `'rad'`
        * **dt** : number, optional
            Time interval for the elevation computation (seconds)
    """
    time   = getTime(time)
    t0     = time.iso.split()[0].replace('-', '/') 
    sname  = '{} -- {}'.format(sources[0], sources[1])
    lname  = loc 
    loc    = getLoc(loc)
    
    # Compute the elevation vs time
    dt    = TimeDelta(dt, format='sec')
    nvals = int((TimeDelta(1, format='jd')/dt).value)
    times = np.zeros( nvals )
    sep  = np.zeros( nvals )
    for i in np.arange(nvals):
        times[i] = time.mjd
        sep[i]  = getSep(sources[0], sources[1], time=time, loc=loc, unit='deg').deg
        time += dt

    # Interpolate
    evst  = interp1d(times, sep, kind='cubic')
    times = np.linspace(times[0], times[-1], 1000) 
    sep  = evst(times)

    # Plot
    fig, ax = plt.subplots()
    xtime = mdates.date2num( Time(times, format='mjd').to_datetime() )
    ax.plot( xtime, sep, label='{} @ {}'.format(sname, lname) )
    ax.axhline(0., linestyle='--', color='black', linewidth=1)
    ymin, ymax = ax.get_ylim()
    ax.xaxis_date()
    date_format = mdates.DateFormatter('%H:%M')
    ax.xaxis.set_major_formatter(date_format)
    ax.xaxis.set_minor_formatter(date_format)
    fig.autofmt_xdate()
    ax.legend()
    ax.set_xlabel('Time ({} UTC)'.format(t0))
    ax.set_ylabel('Separation (deg)')
    plt.show()



#! /usr/bin/python3.5
# -*- coding: utf-8 -*-

"""
Class to read XST NenuFAR data
        by A. Loh
"""

__author__ = 'Alan Loh'
__copyright__ = 'Copyright 2018, nenupy'
__credits__ = ['Alan Loh']
__license__ = 'MIT'
__version__ = '0.0.1'
__maintainer__ = 'Alan Loh'
__email__ = 'alan.loh@obspm.fr'
__status__ = 'WIP'
__all__ = ['XST']


import os
import sys
import glob
import numpy as np

import matplotlib as mpl
from matplotlib import pyplot as plt
from mpl_toolkits import axes_grid1
from mpl_toolkits.axes_grid1 import make_axes_locatable

from astropy.io import fits
from astropy.time import Time, TimeDelta


class XST():
    def __init__(self, obsfile):
        self.obsfile = obsfile
        self.ma1     = 0
        self.ma2     = 0

    def __str__(self):
        toprint  = '\t=== Class XST of nenupy ===\n'
        toprint += '\tList of all current attributes:\n'
        for att in dir(self):
            avoid = ['d', 'maposition', 'marotation']
            if (not att.startswith('_')) & (not any(x.isupper() for x in att)) & (att not in avoid):
                toprint += "%s: %s\n"%(att, getattr(self, att))
        return toprint

    # ================================================================= #
    # ======================== Getter / Setter ======================== #
    @property
    def obsfile(self):
        """ XST observation file,
        """
        return self._obsfile
    @obsfile.setter
    def obsfile(self, o):
        if (o is None) or (o == '') or (os.path.isdir(o)):
            if os.path.isdir(o):
                _opath = os.path.abspath(o)
                bstfiles = glob.glob( os.path.join(_opath, '*XST.fits') )
            else:
                bstfiles = glob.glob('*XST.fits')
            if len(bstfiles) == 0:
                raise IOError("\n\t=== No XST fits file in current directory, specify the file to read. ===")
            elif len(bstfiles) == 1:
                o = os.path.abspath(bstfiles[0])
            else:
                raise IOError("\n\t=== Multiple XST files are not handled yet ===")
        else:
            if not os.path.isfile(o):
                raise IOError("\t=== File {} not found. ===".format(os.path.abspath(o)))
            else:
                o = os.path.abspath(o) 
        self._obsfile = o

        if not self._isXST():
            raise ValueError("\t=== Files might not be XST observaiton ===")
        else:
            self._readXST()

        return

    @property
    def ma1(self):
        """ Selected Mini-Array number 1
        """
        return self._ma1
    @ma1.setter
    def ma1(self, m):
        if m is None:
            m = 0
        if not isinstance(m, int):
            try:
                m = int(m)
            except:
                print("\n\t=== WARNING: Mini-Array index {} not recognized ===".format(m))
                m = 0
        if m in self.miniarrays:
                self._ma1 = m
        else:
            print("\n\t=== WARNING: available Mini-Arrays are {} ===".format(self.miniarrays))
            self._ma1 = 0
        return

    @property
    def ma2(self):
        """ Selected Mini-Array number 2
        """
        return self._ma2
    @ma2.setter
    def ma2(self, m):
        if m is None:
            m = 0
        if not isinstance(m, int):
            try:
                m = int(m)
            except:
                print("\n\t=== WARNING: Mini-Array index {} not recognized ===".format(m))
                m = 0
        if m in self.miniarrays:
                self._ma2 = m
        else:
            print("\n\t=== WARNING: available Mini-Arrays are {} ===".format(self.miniarrays))
            self._ma2 = 0
        return

    # ================================================================= #
    # =========================== Methods ============================= #
    def getData(self, **kwargs):
        """ Make the data selection
            Fill the attributes self.d (data)
        """
        self._evalkwargs(kwargs)
        baseline = np.sort( [self.ma1, self.ma2] )[::-1] 

        # ------ Indices ------
        # These formulas were defined to fit the table of detail_FITS_00_05.pdf
        MA1X_MA2X = baseline[0]*2*(baseline[0]*2+1)/2       + 2*baseline[1]
        MA1X_MA2Y = baseline[0]*2*(baseline[0]*2+1)/2+1     + 2*baseline[1] # it will be the same as next one for auto-corr
        MA1Y_MA2X = (baseline[0]*2+1)*(baseline[0]*2+2)/2   + 2*baseline[1]
        MA1Y_MA2Y = (baseline[0]*2+1)*(baseline[0]*2+2)/2+1 + 2*baseline[1]
        index     = np.array([MA1X_MA2X, MA1X_MA2Y, MA1Y_MA2X, MA1Y_MA2Y]).astype(int)
        data      = fits.getdata(self.obsfile, ext=7, memmap=True)['DATA'][:, :, index] # time, subband, correls
        # Auto-correlation
        if self.ma1 == self.ma2:
            # Index 1 (XY) is wrong for auto-correlation with the above algorithm,
            # doesn't matter, we replace the data by the conjugate of YX which exists.
            data[:, :, 1] = np.conjugate(data[:, :, 2])
        self.d = data
        return

    def plotCorrMat(self, polar='XX'):
        """ Plot a matrix of Cross correlations mean amplitudes
        """
        data = fits.getdata(self.obsfile, ext=7, memmap=True)['DATA'][:, 0, :] 
        data = np.mean(data, axis=0)

        try:
            pol = ['XX', 'XY', 'YX', 'YY'].index(polar.upper())
        except ValueError:
            raise

        amp = np.zeros( (self.allma.size, self.allma.size) )    

        for ant1 in self.allma:
            for ant2 in self.allma[ant1:]:
                baseline = np.sort( [ant1, ant2] )[::-1] # descending order
                # ------ Indices ------
                # These formulas were defined to fit the table of detail_FITS_00_05.pdf (it a kind of triangular series)
                MA1X_MA2X = baseline[0]*2*(baseline[0]*2+1)/2       + 2*baseline[1]
                MA1X_MA2Y = baseline[0]*2*(baseline[0]*2+1)/2+1     + 2*baseline[1] # it will be the same as next one for auto-corr
                MA1Y_MA2X = (baseline[0]*2+1)*(baseline[0]*2+2)/2   + 2*baseline[1]
                MA1Y_MA2Y = (baseline[0]*2+1)*(baseline[0]*2+2)/2+1 + 2*baseline[1]
                index     = np.array([MA1X_MA2X, MA1X_MA2Y, MA1Y_MA2X, MA1Y_MA2Y]).astype(int)
                d         = data[index]
                if ant1 == ant2:
                    # Index 1 (XY) is wrong for auto-correlations with the above algorithm,
                    # doesn't matter, we replace the data by the conjugate of YX which exists anyway.
                    d[1] = np.conjugate(d[2])
                if not (ant1 == ant2):
                    amp[ant1, ant2] = np.absolute(d[pol])
        amp[amp == 0] = np.nan

        def add_colorbar(im, aspect=20, pad_fraction=0.5, **kwargs):
            """Add a vertical color bar to an image plot."""
            divider = axes_grid1.make_axes_locatable(im.axes)
            width = axes_grid1.axes_size.AxesY(im.axes, aspect=1./aspect)
            pad = axes_grid1.axes_size.Fraction(pad_fraction, width)
            current_ax = plt.gca()
            cax = divider.append_axes("right", size=width, pad=pad)
            plt.sca(current_ax)
            return im.axes.figure.colorbar(im, cax=cax, **kwargs)

        fig, ax = plt.subplots(figsize=(10, 10))
        ax.set_title('{} - {}'.format( os.path.basename(self.obsfile), polar.upper() ))
        ax.set_xlabel('MA')
        ax.set_ylabel('MA')
        im = plt.imshow(np.log10(amp), origin='lower', cmap='Spectral', interpolation='nearest')
        ax.grid(color='k', linestyle='-', linewidth=0.5, alpha=0.5)
        # cb = fig.colorbar(im, ax=ax)
        # cb.set_label('log$_{10}$( Amplitude )')
        cb = add_colorbar(im)
        cb.set_label('log$_{10}$( Amplitude )')
        plt.show()
        return        

    def plotData(self, **kwargs):
        """ Plot
        """
        self.getData(**kwargs)
        xstamp   = np.absolute(self.d)
        xstphase = np.angle(self.d, deg=True)

        # ------ Amplitude ------ #
        fig, (ax1, ax2, ax3, ax4) = plt.subplots(4, 1)
        fig.suptitle(r'Amplitude (Baseline MA{}-MA{})'.format(self.ma1, self.ma2), fontsize=15)
        im1 = ax1.imshow(xstamp[:, :, 0].T, origin='lower', aspect='auto', interpolation='nearest')
        ax1.xaxis.set_visible(False)
        ax1.set_title(r'MA{}X-MA{}X'.format(self.ma1, self.ma2), fontsize=10)
        divider = make_axes_locatable(ax1)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(im1, cax=cax)
        ax1.set_ylabel('Freq steps')
        im2 = ax2.imshow(xstamp[:, :, 1].T, origin='lower', aspect='auto', interpolation='nearest')
        ax2.xaxis.set_visible(False)
        ax2.set_title(r'MA{}X-MA{}Y'.format(self.ma1, self.ma2), fontsize=10)
        divider = make_axes_locatable(ax2)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(im2, cax=cax)
        ax2.set_ylabel('Freq steps')
        im3 = ax3.imshow(xstamp[:, :, 2].T, origin='lower', aspect='auto', interpolation='nearest')
        ax3.xaxis.set_visible(False)
        ax3.set_title(r'MA{}Y-MA{}X'.format(self.ma1, self.ma2), fontsize=10)
        divider = make_axes_locatable(ax3)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(im3, cax=cax)
        ax3.set_ylabel('Freq steps')
        im4 = ax4.imshow(xstamp[:, :, 3].T, origin='lower', aspect='auto', interpolation='nearest')
        ax4.set_xlabel('Time steps')
        ax4.set_title(r'MA{}Y-MA{}Y'.format(self.ma1, self.ma2), fontsize=10)
        divider = make_axes_locatable(ax4)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(im4, cax=cax)
        ax4.set_ylabel('Freq steps')
        plt.show()
        plt.close('')

        # ------ Phase ------ #
        fig, (ax1, ax2, ax3, ax4) = plt.subplots(4, 1)
        fig.suptitle(r'Phase (Baseline MA{}-MA{})'.format(self.ma1, self.ma2), fontsize=15)
        im1 = ax1.imshow(xstphase[:, :, 0].T, origin='lower', aspect='auto', interpolation='nearest')
        ax1.xaxis.set_visible(False)
        ax1.set_title(r'MA{}X-MA{}X'.format(self.ma1, self.ma2), fontsize=10)
        ax1.set_ylabel('Freq steps')
        divider = make_axes_locatable(ax1)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(im1, cax=cax)
        im2 = ax2.imshow(xstphase[:, :, 1].T, origin='lower', aspect='auto', interpolation='nearest')
        ax2.xaxis.set_visible(False)
        ax2.set_title(r'MA{}X-MA{}Y'.format(self.ma1, self.ma2), fontsize=10)
        divider = make_axes_locatable(ax2)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(im2, cax=cax)
        ax2.set_ylabel('Freq steps')
        im3 = ax3.imshow(xstphase[:, :, 2].T, origin='lower', aspect='auto', interpolation='nearest')
        ax3.xaxis.set_visible(False)
        ax3.set_title(r'MA{}Y-MA{}X'.format(self.ma1, self.ma2), fontsize=10)
        divider = make_axes_locatable(ax3)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(im3, cax=cax)
        ax3.set_ylabel('Freq steps')
        im4 = ax4.imshow(xstphase[:, :, 3].T, origin='lower', aspect='auto', interpolation='nearest')
        ax4.set_xlabel('Time steps')
        ax4.set_title(r'MA{}Y-MA{}Y'.format(self.ma1, self.ma2), fontsize=10)
        divider = make_axes_locatable(ax4)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(im4, cax=cax)
        ax4.set_ylabel('Freq steps')
        plt.show()
        plt.close('')
        return

    def convertMS(self):
        """ Convert the XST data into a Measurement Set
        """
        return

    # ================================================================= #
    # =========================== Internal ============================ #
    def _isXST(self):
        """ Check that self.obsfile is a proper XST observation
        """
        isXST = True
        with fits.open(self.obsfile, mode='readonly', ignore_missing_end=True, memmap=True) as f:
            if f[0].header['OBJECT'] != 'crosscorrelation Statistics':
                isXST = False
            else:
                pass
        return isXST

    def _readXST(self):
        """ Read XST fits files and fill the class attributes
        """
        self.obsname = os.path.basename( self.obsfile )[:-9]

        with fits.open(self.obsfile, mode='readonly', ignore_missing_end=True, memmap=True) as f:
            head      = f[0].header
            setup_ins = f[1].data
            setup_obs = f[2].data
            setup_ana = f[3].data
            setup_bea = f[4].data
            setup_pan = f[5].data
            setup_pbe = f[6].data

        self.obstart  = Time( head['DATE-OBS'] + 'T' + head['TIME-OBS'] )
        self.obstop   = Time( head['DATE-END'] + 'T' + head['TIME-END'] )
        self.time     = [self.obstart.copy(), self.obstop.copy()]
        self.exposure = self.obstop - self.obstart

        self.allma      = np.squeeze( setup_ins['noMR'])
        self.miniarrays = np.squeeze( setup_ins['noMROn'] )
        self._marot     = np.squeeze( setup_ins['rotation'] )
        self._mapos     = np.squeeze( setup_ins['noPosition'] )
        self._mapos     = self._mapos.reshape( int(self._mapos.size/3), 3 )
        self._pols      = np.squeeze( setup_ins['spol'] )

        self.abeams     = setup_ana['NoAnaBeam']
        self._antlist   = np.array( [ np.array(eval(i)) - 1 for i in setup_ana['Antlist'] ])

        self.dbeams     = setup_bea['noBeam']
        self._digi2ana  = setup_bea['NoAnaBeam']
        self._bletlist  = setup_bea['BeamletList']
        self._freqs     = setup_bea['freqList']

        self._pointana  = setup_pan['noAnaBeam']
        self._azlistana = setup_pan['AZ']
        self._ellistana = setup_pan['EL']
        self._pointanat = Time(np.array([setup_pan['timestamp'][self._pointana==i] for i in self.abeams]))
        
        self._pointdig  = setup_pbe['noBeam']
        self._azlistdig = setup_pbe['AZ']
        self._ellistdig = setup_pbe['EL']    
        self._pointdigt = Time(setup_pbe['timestamp'])
        return

    def _evalkwargs(self, kwargs):
        for key, value in kwargs.items():
            if   key == 'ma1': self.ma1  = value
            elif key == 'ma2': self.ma2  = value
            else:
                pass
        return





#! /usr/bin/python3.6
# -*- coding: utf-8 -*-

"""
Class to do fancy progressbars
        by G. Schworer
https://github.com/ceyzeriat/patiencebar
"""

from sys import stdout as _stdout

class ProgressBar(object):
    """
    Provides a terminal-friendly single-thread progress bar
    Args:
      * valmax (float): the finish value of the progress bar. Default is 100.
      * barsize (int >0): the size of the bar in the opened terminal. If ``None``, the bar will automatically fit the width of the window.
      * title (str): the title, printed one line above the progress bar
      * bar (bool): whether the bar should be displayed or not. If ``False``, only the text given at each :func:`update` will be printed
      * up_every (int [0-100]): if ``bar`` is ``True``, the progress bar will be updated every ``up_every`` percent of progress. Setting ``up_every`` = 0 updates the progress bar at each :func:`update`
    >>> import patiencebar as PB
    >>> n_calc = 34
    >>> pb = PB.ProgressBar(valmax=n_calc, barsize=50, title="Test bar")
    >>> for i in range(n_calc):
    >>>     do_stuff()
    >>>     pb.update()
    """
    def __init__(self, valmax=100, barsize=None, title=None, bar=True, up_every=2):
        self.reset(valmax=valmax, barsize=barsize, title=title, bar=bar, up_every=up_every)

    @property
    def valmax(self):
        return self._valmax
    @valmax.setter
    def valmax(self, value):
        raise AttributeError("Read-only")

    @property
    def barsize(self):
        return self._barsize
    @barsize.setter
    def barsize(self, value):
        raise AttributeError("Read-only")

    @property
    def title(self):
        return self._title
    @title.setter
    def title(self, value):
        raise AttributeError("Read-only")

    @property
    def running(self):
        return self._running
    @running.setter
    def running(self, value):
        raise AttributeError("Read-only")

    @property
    def bar(self):
        return self._bar
    @bar.setter
    def bar(self, value):
        raise AttributeError("Read-only")

    @property
    def up_every(self):
        return self._up_every
    @up_every.setter
    def up_every(self, value):
        raise AttributeError("Read-only")


    def reset(self, valmax=None, barsize=None, title=None, bar=None, up_every=None):
        """
        Resets the progress bar with initialization values, unless new values are given
        Args:
          * valmax (float): the finish value of the progress bar. Default is 100.
          * barsize (int >0): the size of the bar in the opened terminal. If ``None``, the bar will automatically fit the width of the window.
          * title (str): the title, printed one line above the progress bar.
          * bar (bool): whether the bar should be displayed or not. If ``False``, only the text given at each :func:`update` will be printed.
          * up_every (int [0-100]): if ``bar`` is ``True``, the progress bar will be updated every ``up_every`` percent of progress. Setting ``up_every`` = 0 updates the progress bar at each :func:`update`.
        >>> n_calc = 34
        >>> pb = PB.Patiencebar(valmax=n_calc, barsize=50, title="Test bar")
        >>> for i in range(n_calc):
        >>>     do_stuff()
        >>>     pb.update()
        >>> pb.reset(title="Second trial", barsize=70)
        >>> for i in range(n_calc):
        >>>     do_stuff()
        >>>     pb.update()
        """
        self.win_width = 50 #_get_terminal_size()[1]
        self.step = 0
        self._nextup = 0
        self._title_written = False
        self._running = False
        if valmax is not None:
            self._valmax = float(valmax)
        else:
            self._valmax = float(getattr(self, 'valmax', 100))
        if barsize is not None:
            self._barsize = int(barsize)
        else:
            self._barsize = int(getattr(self, 'barsize', self.win_width-8))
        if title is not None:
            self._title = str(title)
        elif getattr(self, 'title', None) is not None:
            self._title = str(getattr(self, 'title'))
        else:
            self._title = None
        if bar is not None:
            self._bar = bool(bar)
        else:
            self._bar = bool(getattr(self, 'bar', True))
        if up_every is not None:
            self._up_every = int(min(100,max(0,up_every)))
        else:
            self._up_every = int(getattr(self, 'up_every', 2))

    
    def update(self, step=None):
        """
        Updates the progress bar to a newer value
        Args:
          * step (None): adds 1 to the progress of the bar, where ``valmax`` is the finish value.
          * step (float): sets the progress of the bar to the ``step`` value, where ``valmax`` is the finish value.
          * step (str): displays ``step`` on a new line. For this to work, ``bar`` must be ``False`` (no progress bar displayed) otherwise the update instruction is ignored.
        """
        self._doupdate(step=step)
        
    def _doupdate(self, step=None):
        # initialization
        if not self._title_written:
            if self.title is not None: print(self.title)
            self._title_written = True
            self._running = True # in case it was not yet done
        if self.bar: # if display progressbar
            if step is None: # no step given
                self.step += 1
            else: # step given              
                try: # make sure we got some number
                    self.step = min(float(step), self.valmax)
                except: # if not.. WTF.. don't change anything
                    return
            # render bar
            perc = int(self.step*100./self.valmax)
            if perc < self._nextup: return
            self._nextup = min(100, perc+self._up_every)
            bar = int(perc/100.*self.barsize)
            spacing = (self.win_width-self.barsize-8)/2.
            out = '\r%s[%s%s] %3d %%%s' % (' ' * int(spacing), '=' * bar, ' ' * int(self.barsize - bar), perc, ' ' * int(spacing+0.5))
            _stdout.write(out)
            _stdout.flush()
        else: # if no bar display
            if step is None: step = 'tic'
            print(str(step))
            self.step += 1
        if self.step >= self.valmax: # finished
            print("\r")
            self._running = False            